#!/usr/bin/python
from __future__ import print_function
####################################################################
# reads SmartMeter information from a file and uploads to Domoticz
# 
# version 1.0 / January 2016 / paulvha
# initial version
# 
#################################################################### 

import sys
import json
import urllib2
import re
import time
import datetime
import os, argparse

# constants
DISVERS = "version 1.0"				# display version
UDEBUG = False
TIME_OUT = 120						# Timeout for input file (seconds)

# To print in color
REDSTR = u"\033[1;31m{0}\033[00m"
GRNSTR = u"\033[1;92m{0}\033[00m"
BLUESTR = u"\033[1;96m{0}\033[00m"
YELSTR = u"\033[1;93m{0}\033[00m"

# variables
dom_url = "127.0.0.1"				# -u
dom_port = "8080"					# -p
xid_el = 0							# -e
xid_gas = xid_el + 1				# -g
meter_file = "/tmp/meterinput"		# -f
repeat_cnt = 0 						# -r
interval = 60						# -i
display_info = False				# -d
quiet = False						# -q
gas_is_on_next_line = False			# -L

# for gas update
last_gas_value = 0
l_gas_t = 0

MTR_USAGE_PEAK = 0
MTR_USAGE_LOW = 1
MTR_GEN_PEAK = 2
MTR_GEN_LOW = 3
MTR_USAGE_ACT = 4
MTR_GEN_ACT = 5
MTR_GAS = 6 
MTR_GAS_NXT = 7

# Matchlist for OBIS to sent to Domoticz
OBIS_table =  [[MTR_USAGE_PEAK,{"OBIS": "1-0:1.8.1", "start": '(', "last": '*', "value": 0}],	
 [MTR_USAGE_LOW,{"OBIS": "1-0:1.8.2",  "start": '(', "last": '*', "value": 0}],	
 [MTR_GEN_PEAK,{"OBIS": "1-0:2.8.1",  "start": '(', "last": '*', "value": 0}],	
 [MTR_GEN_LOW,{"OBIS": "1-0:2.8.2",  "start": '(', "last": '*', "value": 0}],	
 [MTR_USAGE_ACT,{"OBIS": "1-0:1.7.0",  "start": '(', "last": '*', "value": 0}],	
 [MTR_GEN_ACT,{"OBIS": "1-0:2.7.0",  "start": '(', "last": '*', "value": 0}],
 [MTR_GAS,{"OBIS": "0-1:24.2.1",  "start": '(', "last": '*', "value": 0}],	
 [MTR_GAS_NXT,{"OBIS": "NO MATCH",  "start": '(', "last": ')', "value": 0}]]	

# print to STDERR (implement future/python 3 comp)
# white, green, blue, yellow and red (= default) are supported
def printstderr(message,dcol = "red"):
	
	if quiet:	return 
		
	message = "domo_update: " + message
	
	if dcol == "red": message=REDSTR.format(message)
	elif dcol == "green": message=GRNSTR.format(message)
	elif dcol == "blue": message=BLUESTR.format(message)
	elif dcol == "yellow": message=YELSTR.format(message)
	
	print(message, file=sys.stderr)

def display_error_exit(message):
	printstderr(message)
	sys.exit(1)
	
# Displays message to the user
# red, green, blue, yellow and white (= default) are supported

def display_mes(message,dcol = "white"):

	if quiet or display_info == False:	return 

	if dcol == "red": message = REDSTR.format(message)
	elif dcol == "green": message = GRNSTR.format(message)
	elif dcol == "blue": message = BLUESTR.format(message)
	elif dcol == "yellow": message = YELSTR.format(message)

	print(message)
		
# update an OBIS selection
def set_obis(entry, code):
	global OBIS_table

	OBIS_table[entry][1]['OBIS'] = str(code)
		
	if (UDEBUG): printstderr("set_obis: {}".format(OBIS_table),"yellow")	

# read meter input from file
def read_file(meter_file):
	input_buffer = ""
	ii = True
	tt = time.time()
	
	while ii:
		if os.access(meter_file, os.R_OK) == 0:	 	# is file there ?
			if (UDEBUG): printstderr("waiting for {} (retry in 10 seconds)".format(meter_file),"yellow")
		
			if time.time() - tt > TIME_OUT:
				display_error_exit("Error: timeout waiting {}".format(meter_file))
				
			time.sleep(10)
		else:
			ii = False	
	
	ret = os.stat(meter_file) 						# if empty file
	
	if ret.st_size == 0:
		display_error_exit("Error: {} is empty".format(meter_file))
		
	try:
		readfile = open(meter_file,'rb')				

		if (readfile):								# read the data
			input_buffer = readfile.read()

			readfile.close()
			
			display_mes(input_buffer)				# in case -d 
			
			return(input_buffer)					# succesfull reading
		else:
			display_error_exit("Error: during reading {}".format(meter_file))	
			
	except Exception, e:
		display_error_exit("Error: exception {} during handling {}".format(e, meter_file))

# find the OBIS selection in the input_buffer
def select_data(input_buffer):
	global OBIS_table
	
	out_buffer = ""
	f_count = 0
	gas_found = False	

	num = len(OBIS_table)
	input_lines = input_buffer.splitlines()

	for kk in range(0,len(input_lines)):			# for each line

		i_line = input_lines[kk]
		
		if i_line == "": continue					# empty line
		
		jj = 0			
		
		while jj < num:								# for each OBIS selection

			obis_code = OBIS_table[jj][1]['OBIS']
				
			if gas_found:							# gas was found 
				jj = MTR_GAS_NXT					# now on next line
				obis_code = s_obis
				gas_found = False
			
			elif i_line.find(obis_code) < 0:		# else try to match
				jj += 1								# no match: next OBIS	
				continue

			if (UDEBUG): printstderr((obis_code + " " + i_line),"yellow")
			
			if OBIS_table[jj][0] == MTR_GAS:		# if gas was found
				if gas_is_on_next_line:				# & -L option
					gas_found = True
					s_obis = obis_code
					break

			f_sign = OBIS_table[jj][1]['start']		# start character
			l_sign = OBIS_table[jj][1]['last']		# ending character

			f_pos = i_line.rfind(f_sign) + 1		# get start position
			l_pos = i_line.find(l_sign, f_pos)		# get last position

			if f_pos < 0 or l_pos < 0:				# found ?
				continue
		
			value = i_line[f_pos:l_pos]				# get value

			OBIS_table[jj][1]['value'] = value		# store value		

			out_buffer += "{}\t{}\n".format(obis_code, value)

			f_count += 1

			break

	display_mes(out_buffer)							# in case -d

	if f_count != num -1 :
		display_error_exit("Could not locate all the OBIS selection.\n {}".format(out_buffer))	


# Class to connect to Domoticz

class dom_class():
    
    def __init__(self, url, port):
		 
       if url[0] == 'h':
		   self.baseurl = "{}:{}".format(url,port)
       else:
           self.baseurl = "http://{}:{}".format(url,port)
       
       if (UDEBUG): printstderr("dom_class.init {}".format(self.baseurl),"yellow")
        
    def send_info(self, url):
        if (UDEBUG): printstderr("dom_class.send_info {}".format(url),"yellow")
        req = urllib2.Request(url)
        return urllib2.urlopen(req, timeout=5)

    def update_device(self, xid, values):
        """
        update the device information.
        """		
        url = "%s/json.htm?type=command&param=udevice&idx=%s&nvalue=0&svalue=%s" % (self.baseurl, xid, values)
        return json.load(self.send_info(url))


# update electricity information
def update_elec(dom,idx):

	el1 = int(float(OBIS_table[MTR_USAGE_PEAK][1]['value']) * 1000)
	el2 = int(float(OBIS_table[MTR_USAGE_LOW][1]['value']) * 1000)
	el3 = int(float(OBIS_table[MTR_GEN_PEAK][1]['value']) * 1000)
	el4 = int(float(OBIS_table[MTR_GEN_LOW][1]['value']) * 1000)
	el5 = int(float(OBIS_table[MTR_USAGE_ACT][1]['value']) * 1000)
	el6 = int(float(OBIS_table[MTR_GEN_ACT][1]['value']) * 1000)
	
	values = "{};{};{};{};{};{}".format(el1, el2, el3, el4, el5,el6)
	
	if (UDEBUG): printstderr("electricity values {}".format(values),"yellow")
	
	ret = dom.update_device(idx, values)
	
	if ret['status'] != 'OK':
		display_error_exit("Error: during update electricity")
	
	return 0

# update gas information
def update_gas(dom,idx):
	global last_gas_value, l_gas_t
	
	if gas_is_on_next_line:				# in case -L option
		value = OBIS_table[MTR_GAS_NXT][1]['value'] 
	else:
		value = OBIS_table[MTR_GAS][1]['value'] 	

	c_time = time.time()
										# only with change or 10 min	
	
	if value != last_gas_value or c_time - l_gas_t > 600:
		
		last_gas_value = value
		l_gas_t = c_time
		
		values = "{}".format(int(float(value) * 1000))
		
		if (UDEBUG): printstderr("Gas value {}".format(values),"yellow")
		
		ret = dom.update_device(idx, values)
		
		if ret['status'] != 'OK':
			display_error_exit ("Error: during update gas")
	
	return 0

## program start here ##

parser = argparse.ArgumentParser(description="update Domoticz with SmartMeter information",
 epilog = "you must provide electricity ID. (from devices overview in Domoticz)",
 prog = "domo_update")

parser.add_argument('--version', action='version', version=DISVERS)	
parser.add_argument("-d","--display", help="display info read & sent", action="store_true")
parser.add_argument("-f","--file", help="input file (default: /tmp/meterinput)")
parser.add_argument("-i","--wait_time",help="minimum amount of seconds between posting (default: 60)", type = int)
parser.add_argument("-e","--eidx", help="electricity IDX", type = int, required=True)
parser.add_argument("-g","--gidx", help="Gas IDX (default: eidx + 1)", type = int)
parser.add_argument("-p","--port", help="domoticz port (default: 8080)", type = int)
parser.add_argument("-u","--url", help="device to use (default: 127:0:0:1)")
parser.add_argument("-q","--quiet", help="No message will be displayed", action="store_true")
parser.add_argument("-r","--rpt",help="repeat count (default: 0 = continues)", type = int)
parser.add_argument("-D", "--Dp", help="OBIS for usage day or peak rate (default: 1-0:1.8.1)")	
parser.add_argument("-N", "--Do", help="OBIS for usage night or low rate (default: 1-0:1.8.2)")	
parser.add_argument("-P", "--Gp", help="OBIS for generated day or peak rate (default: 1-0:2.8.1)")
parser.add_argument("-O", "--Go", help="OBIS for generated night low rate (default: 1-0:2.8.2)")
parser.add_argument("-A", "--Au", help="OBIS for actual usage (default: 1-0:1.7.0)")	
parser.add_argument("-R", "--Ag", help="OBIS for actual generated (default: 1-0:2.7.0)")
parser.add_argument("-G", "--Ga", help="OBIS for GAS usage (default: 0-1:24.2.0)")					
parser.add_argument("-L", "--next", help="GAS usage is on the line following -G OBIS.", action="store_true")

args = parser.parse_args()

if args.display :	display_info = True				# display information read/sent
elif args.quiet	:	quiet = True					# no display, no error messages etc
		
if args.wait_time:	interval = args.wait_time		# seconds to wait in between
if args.rpt		:	repeat_cnt = args.rpt			# how many repeat (support to start from crontab)
if args.file 	:	meter_file = args.file			# file to read
if args.url 	:	dom_url = args.url				# url to use
if args.port 	:	dom_port = args.port			# port to use	
	
if args.eidx :										# Electricity id must be provided
	idx_el = args.eidx
	idx_gas =  idx_el + 1

if args.gidx: 		idx_gas = args.gidx				# new gas_id

if args.Dp	:	set_obis(MTR_USAGE_PEAK, args.Dp)	# new OBIS ?
if args.Do	:	set_obis(MTR_USAGE_LOW, args.Do)
if args.Gp	:	set_obis(MTR_GEN_PEAK, args.Gp)
if args.Go	:	set_obis(MTR_GEN_LOW, args.Go)
if args.Au	:	set_obis(MTR_USAGE_ACT, args.Au)
if args.Ag	:	set_obis(MTR_GEN_ACT, args.Ag)
if args.Ga	:	set_obis(MTR_GAS, args.Ga)
if args.next:	gas_is_on_next_line = True

# set class
dom = dom_class(dom_url, dom_port)

while True:

	input_buffer = read_file(meter_file)
								
	select_data(input_buffer)

	update_elec(dom,idx_el)

	update_gas(dom,idx_gas)

	if repeat_cnt > 0:						# repeat ?(0 = endless) 
		repeat_cnt -= 1						# subtract last pass
		
		if (UDEBUG): printstderr("repeat counter {}".format(repeat_cnt),"yellow")	
		
		if repeat_cnt == 0:					# we are done ?
			sys.exit(0)
			
	if (UDEBUG): printstderr("sleep for {} seconds\n".format(interval),"yellow")
	
	time.sleep(interval)					# wait
	
	os.remove(meter_file)					# remove current to get new
